/// <reference types="node" />
/// <reference types="bn.js" />
import { BN } from 'ethereumjs-util';
import { Block, BlockHeader } from '@ethereumjs/block';
import Ethash from '@ethereumjs/ethash';
import Common from '@ethereumjs/common';
import { DBManager, DBOp } from './dbManager';
import type { LevelUp } from 'levelup';
declare type OnBlock = (block: Block, reorg: boolean) => Promise<void> | void;
export interface BlockchainInterface {
    /**
     * Adds a block to the blockchain.
     *
     * @param block - The block to be added to the blockchain.
     * @param isGenesis - True if block is the genesis block.
     */
    putBlock(block: Block, isGenesis?: boolean): Promise<void>;
    /**
     * Deletes a block from the blockchain. All child blocks in the chain are deleted and any
     * encountered heads are set to the parent block.
     *
     * @param blockHash - The hash of the block to be deleted
     */
    delBlock(blockHash: Buffer): Promise<void>;
    /**
     * Returns a block by its hash or number.
     */
    getBlock(blockId: Buffer | number | BN): Promise<Block | null>;
    /**
     * Iterates through blocks starting at the specified iterator head and calls the onBlock function
     * on each block.
     *
     * @param name - Name of the state root head
     * @param onBlock - Function called on each block with params (block: Block, reorg: boolean)
     */
    iterator(name: string, onBlock: OnBlock): Promise<void>;
}
/**
 * This are the options that the Blockchain constructor can receive.
 */
export interface BlockchainOptions {
    /**
     * Specify the chain and hardfork by passing a Common instance.
     *
     * If not provided this defaults to chain `mainnet` and hardfork `chainstart`
     *
     */
    common?: Common;
    /**
     * Database to store blocks and metadata. Should be an abstract-leveldown compliant store.
     */
    db?: LevelUp;
    /**
     * This flags indicates if Proof-of-work should be validated. Defaults to `true`.
     */
    validatePow?: boolean;
    /**
     * This flags indicates if blocks should be validated. See Block#validate for details. If
     * `validate` is provided, this option takes its value. If neither `validate` nor this option are
     * provided, it defaults to `true`.
     */
    validateBlocks?: boolean;
}
/**
 * This class stores and interacts with blocks.
 */
export default class Blockchain implements BlockchainInterface {
    db: LevelUp;
    dbManager: DBManager;
    ethash?: Ethash;
    private _genesis?;
    private _headBlock?;
    private _headHeader?;
    private _heads;
    private _staleHeadBlock;
    private _staleHeads;
    private _initDone;
    private _lock;
    private _common;
    private readonly _validatePow;
    private readonly _validateBlocks;
    /**
     * Creates new Blockchain object
     *
     * @param opts - An object with the options that this constructor takes. See [[BlockchainOptions]].
     */
    constructor(opts?: BlockchainOptions);
    /**
     * Returns an object with metadata about the Blockchain. It's defined for backwards compatibility.
     */
    get meta(): {
        rawHead: Buffer | undefined;
        heads: {
            [key: string]: Buffer;
        };
        genesis: Buffer | undefined;
    };
    /**
     * Fetches the meta info about the blockchain from the db. Meta info contains
     * hashes of the headerchain head, blockchain head, genesis block and iterator
     * heads.
     *
     * @hidden
     */
    _init(): Promise<void>;
    /**
     * Sets the default genesis block
     *
     * @hidden
     */
    _setCanonicalGenesisBlock(): Promise<void>;
    /**
     * Puts the genesis block in the database
     *
     * @param genesis - The genesis block to be added
     */
    putGenesis(genesis: Block): Promise<void>;
    /**
     * Returns the specified iterator head.
     *
     * @param name - Optional name of the state root head (default: 'vm')
     */
    getHead(name?: string): Promise<Block>;
    /**
     * Returns the latest header in the canonical chain.
     */
    getLatestHeader(): Promise<BlockHeader>;
    /**
     * Returns the latest full block in the canonical chain.
     */
    getLatestBlock(): Promise<Block>;
    /**
     * Adds many blocks to the blockchain.
     *
     * @param blocks - The blocks to be added to the blockchain
     */
    putBlocks(blocks: Block[]): Promise<void>;
    /**
     * Adds a block to the blockchain.
     *
     * @param block - The block to be added to the blockchain
     */
    putBlock(block: Block, isGenesis?: boolean): Promise<void>;
    /**
     * Adds many headers to the blockchain.
     *
     * @param headers - The headers to be added to the blockchain
     */
    putHeaders(headers: Array<any>): Promise<void>;
    /**
     * Adds a header to the blockchain.
     *
     * @param header - The header to be added to the blockchain
     */
    putHeader(header: BlockHeader): Promise<void>;
    /**
     * @hidden
     */
    _putBlockOrHeader(item: Block | BlockHeader, isGenesis?: boolean): Promise<void>;
    /**
     * Gets a block by its hash.
     *
     * @param blockId - The block's hash or number
     */
    getBlock(blockId: Buffer | number | BN): Promise<Block>;
    /**
     * @hidden
     */
    _getBlock(blockId: Buffer | number | BN): Promise<Block>;
    /**
     * Looks up many blocks relative to blockId
     *
     * @param blockId - The block's hash or number
     * @param maxBlocks - Max number of blocks to return
     * @param skip - Number of blocks to skip apart
     * @param reverse - Fetch blocks in reverse
     */
    getBlocks(blockId: Buffer | BN | number, maxBlocks: number, skip: number, reverse: boolean): Promise<Block[]>;
    /**
     * Given an ordered array, returns an array of hashes that are not in the blockchain yet.
     *
     * @param hashes - Ordered array of hashes
     */
    selectNeededHashes(hashes: Array<Buffer>): Promise<Buffer[]>;
    /**
     * @hidden
     */
    _saveHeadOps(): DBOp[];
    /**
     * @hidden
     */
    _saveHeads(): Promise<void>;
    /**
     * Delete canonical number assignments for specified number and above
     *
     * @hidden
     */
    _deleteStaleAssignments(number: BN, headHash: Buffer, ops: DBOp[]): Promise<void>;
    /**
     * Overwrites stale canonical number assignments.
     *
     * @hidden
     */
    _rebuildCanonical(header: BlockHeader, ops: DBOp[]): Promise<void>;
    /**
     * Deletes a block from the blockchain. All child blocks in the chain are deleted and any
     * encountered heads are set to the parent block.
     *
     * @param blockHash - The hash of the block to be deleted
     */
    delBlock(blockHash: Buffer): Promise<void>;
    /**
     * @hidden
     */
    _delBlock(blockHash: Buffer): Promise<void>;
    /**
     * @hidden
     */
    _delChild(hash: Buffer, number: BN, headHash: Buffer | null, ops: DBOp[]): Promise<void>;
    /**
     * Iterates through blocks starting at the specified iterator head and calls the onBlock function
     * on each block. The current location of an iterator head can be retrieved using the `getHead()`
     * method.
     *
     * @param name - Name of the state root head
     * @param onBlock - Function called on each block with params (block, reorg)
     */
    iterator(name: string, onBlock: OnBlock): Promise<void>;
    /**
     * @hidden
     */
    _iterator(name: string, onBlock: OnBlock): Promise<void>;
    /**
     * Gets a header by hash and number. Header can exist outside the canonical chain
     *
     * @hidden
     */
    _getHeader(hash: Buffer, number?: BN): Promise<BlockHeader>;
    /**
     * Gets a header by number. Header must be in the canonical chain
     *
     * @hidden
     */
    _getCanonicalHeader(number: BN): Promise<BlockHeader>;
    /**
     * Gets total difficulty for a block specified by hash and number
     *
     * @hidden
     */
    _getTd(hash: Buffer, number?: BN): Promise<BN>;
}
export {};
